(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_nonempty_set.ml,v 1.15 2013-12-02 15:28:38 fclement Exp $ *)

(* Matrix slot nonempty sets. *)

module Nat = Unatural;;

module Midx = Umatrix_index;;
module Iidx = Uindex_index;;
module Muniv = Umatrix_universe;;
module Iuniv = Uindex_universe;;

let mns_invalid_arg = Upervasives.invalid_arg "Umatrix_nonempty_set";;

type size = Muniv.size;;
type element = Muniv.element;;
          (* = Nat.t * Nat.t *)
type uindex = Muniv.index;;
         (* = element Iuniv.element
            = (Nat.t * Nat.t) Uindex.t *)
type index = element Iuniv.index;;
        (* = (Nat.t * Nat.t) Uindex.t Uindex.t *)

type t =
  {
   universe : Muniv.t;
   uindices : element Iuniv.t;
  }
;;

(* Construction functions. *)

let full mu =
  let max_uidx = Midx.of_natural (Muniv.cardinal mu) in
  let uindices = Iuniv.of_nonempty_range Midx._0_ max_uidx in
  { universe = mu; uindices = uindices; }
;;

let make sz = full (Muniv.make sz);;

let unsafe_of_list1 mu uindices =
  { universe = mu; uindices = Iuniv.of_list1 uindices; }
;;

(* Access functions. *)

let cardinal mns = Iuniv.cardinal mns.uindices;;

let uindex mns = Iuniv.element mns.uindices;;

let choice mns =
  let idx0 = Iidx._0_ in
  let uidx0 = uindex mns idx0 in
  (idx0, uidx0)
;;

let element mns idx =
  let uidx = uindex mns idx in
  Muniv.element mns.universe uidx
;;

(* Iterators. *)

let iter f mns = Iuniv.iter f mns.uindices;;

let sub_iter f mns = Iuniv.sub_iter f mns.uindices;;

let fold_left f accu mns = Iuniv.fold_left f accu mns.uindices;;

let full_fold_left f accu ({ universe = mu; _; } as mns) =
  let len = cardinal mns in
  let (_, nj) = Muniv.size mu in
  let rec loop accu idx uidx (i, j as x) =
    if Iidx.to_natural idx = len then accu else
    let accu = f accu idx uidx x in
    let idx = Iidx.succ idx in
    let uidx = Midx.succ uidx in
    let (i, j) =
      if j = Nat.pred nj then (Nat.succ i, Nat._0_) else (i, Nat.succ j) in
    loop accu idx uidx (i, j) in
  loop accu Iidx._0_ Midx._0_ (Nat._0_, Nat._0_)
;;

let full_iter f = full_fold_left (fun _ idx uidx x -> f idx uidx x) ();;

(* Nonempty subsets. *)

type nonempty_subset =
  {
   support : t;
   indices : index array;
   nonempty_set : t;
  }
;;

let same_universe mns1 mns2 = mns1.universe == mns2.universe;;

let check_same_universe fun_name mns1 mns2 =
  if not (same_universe mns1 mns2) then
    mns_invalid_arg fun_name "universe clash"
;;

let is_full mns = cardinal mns = Muniv.cardinal mns.universe;;

let check_full fun_name mns =
  if not (is_full mns) then
    mns_invalid_arg fun_name "nonempty set is not full"
;;

let check_same_full fun_name mns1 mns2 =
  check_same_universe fun_name mns1 mns2;
  check_full fun_name mns1;
  check_full fun_name mns2
;;

let (full_index_of_uindex : uindex -> index) =
  fun uidx -> Iidx.of_natural (Midx.to_natural uidx)
;;

(* We use here the fact that the full set, containing all the elements of the
  universe, is numbered in the same order as the universe.
  Hence, the use of magic function full_index_of_uindex.
  Should efficiency be an issue, consider using Obj.magic on the whole array of
  indices. *)
let subset_of_full full mns =
  check_full "subset_of_full" full;
  check_same_universe "subset_of_full" full mns;
  let indices =
    Array.map full_index_of_uindex (Iuniv.elements mns.uindices) in
  { support = full; indices = indices; nonempty_set = mns; }
;;

type dichotomy = nonempty_subset * nonempty_subset;;

let dichotomy p mns =
  let dicho (tl, fl) idx uidx =
    if p idx uidx then (idx :: tl, fl) else (tl, idx :: fl) in
  let (tl, fl) =
    match fold_left dicho ([], []) mns with
    | ([], _) | (_, []) -> mns_invalid_arg "dichotomy" "empty subset"
    | (tx :: tl, fx :: fl) -> ((tx, tl), (fx, fl)) in
  let utl = Ulist1.rev_map (uindex mns) tl
  and ufl = Ulist1.rev_map (uindex mns) fl in
  let tmns = unsafe_of_list1 mns.universe utl
  and fmns = unsafe_of_list1 mns.universe ufl in
  ({ support = mns; indices = Uarray.of_list1 tl; nonempty_set = tmns; },
   { support = mns; indices = Uarray.of_list1 fl; nonempty_set = fmns; })
;;
