(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_quantity.ml,v 1.21 2013-12-02 15:28:38 fclement Exp $ *)

module Nat = Unatural;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;
module Triple = Tuple.Triple;;
module Config = Configuration;;

let sign_colormap thres i =
  let thres = Nat.to_int thres in
  if i < - thres then 0 else
  if i > thres then 255 else
  if i < 0 then 63 else
  if i > 0 then 191 else
  127
;;

let dq_print_debug_no_arg = Config.print_debug_no_arg "Display_quantity";;

open Segmentation_types;;
open Image;;
open Segmentation_spec;;
open Display_types;;
open Display_window;;
open Display_quantity_rc;;

module Make =
  functor (Scal : SCALAR) ->
  functor (St : STATE) ->
struct
  type scalar = Scal.t;;
  type other_components = St.other_components;;
  
  let ppf = Config.get_formatter ();;
  
  (* Display_quantity. *)
  
  let fk_dq =
    {
     print_pixmaps = fake_print_pixmaps;
     alter_pixmaps = fake_alter_pixmaps;
     draw_pixmaps = fake_draw_pixmaps;
    }
  ;;
  
  let mk_dq q cf sc gk rgb_dir d_sz =
    let pxms = make_pixmaps q gk d_sz in
    let print_pixmaps = make_print_pixmaps q pxms gk in
    let alter_pixmaps = make_alter_pixmaps q cf sc pxms gk in
    let lls = make_lower_left_corners q gk rgb_dir d_sz margin in
    let draw_pixmaps = make_draw_pixmaps q pxms lls gk in
    {
     print_pixmaps = print_pixmaps;
     alter_pixmaps = alter_pixmaps;
     draw_pixmaps = draw_pixmaps;
    }
  ;;
  
  (* Display_quantities. *)
  
  let fk_dqs = { dq_param = fk_dq; dq_data = fk_dq; dq_grad = fk_dq; };;
  
  let { Uaref.get = get_display_quantities; set = set_display_quantities; } =
    Uaref.make_aref_unique "display quantities" fk_dqs
  ;;
  
  let colormap_filter thres = function
    | Param | Data -> Upervasives.id
    | Grad -> sign_colormap thres
  ;;
  
  let scalar_combinator = function
    | Param -> Upervasives.proj2_2
    | Data -> Upervasives.proj2_1
    | Grad -> Scal.( -. )
  ;;
  
  let make_display_quantities thres gk rgb_dir d_sz =
    let cf = colormap_filter thres in
    let sc = scalar_combinator in
    let mk_dq q = mk_dq q (cf q) (sc q) in
    {
     dq_param = mk_dq Param gk rgb_dir d_sz;
     dq_data = mk_dq Data gk rgb_dir d_sz;
     dq_grad = mk_dq Grad gk rgb_dir d_sz;
    }
  ;;
  
  let get_dq dqs = function
    | Param -> dqs.dq_param
    | Data -> dqs.dq_data
    | Grad -> dqs.dq_grad
  ;;
  
  (* Setup. *)
  
  let display_quantities_setup thres gk rgb_dir d_sz =
    set_display_quantities (make_display_quantities thres gk rgb_dir d_sz);
  ;;
  
  let setup thres gk min_w_sz max_w_sz sz =
    window_setup gk min_w_sz max_w_sz sz;
    let { rgb_direction = rgb_dir;
          display_image_size = d_sz; _; } = get_graphics_elements () in
    display_quantities_setup thres gk rgb_dir d_sz;
  ;;
  
  (* Drawing shortcuts. *)
  
  let print_apply = print_apply dq_print_debug_no_arg;;
  
  let zonation_split { gimage_data = gdata; cimage = (_j, pim); _; } =
    (gdata, pim)
  ;;
  let triple_split abt = (Triple.map fst abt, Triple.map snd abt);;
  let triple_zonation_split zt = triple_split (Triple.map zonation_split zt);;
  
  let alter_quantity q iter stct gdatat pimt =
    let dqs = get_display_quantities () in
    let { print_pixmaps = prts; alter_pixmaps = alts; _; } = get_dq dqs q in
    alter_apply iter stct gdatat pimt alts;
    print_apply ppf prts;
  ;;
  
  let draw_quantity q =
    let dqs = get_display_quantities () in
    let { draw_pixmaps = drws; _; } = get_dq dqs q in
    draw_apply drws;
  ;;
  
  let alter_and_draw_quantity iter q zt =
    let stct = scalar_to_channels Scal.to_int in
    let (gdatat, pimt) = triple_zonation_split zt in
    alter_quantity q iter stct gdatat pimt;
    draw_quantity q;
  ;;
  
  let iter_set = pimage_iter_value
  and iter_upd = pimage_iter_to_do
  ;;
  
  let set_quantity_and_draw = alter_and_draw_quantity iter_set
  and update_quantity_and_draw = alter_and_draw_quantity iter_upd
  ;;
end
;;
