(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2013 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_quantity_rc.ml,v 1.14 2013-12-15 18:04:28 fclement Exp $ *)

module Nat = Unatural;;
module Prational = Upositive_rational;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Triple = Tuple.Triple;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _2_ = Nat._2_
and _3_ = Nat._3_
;;

let natural_axpby a x b y = Nat.add (Nat.mul a x) (Nat.mul b y);;

(* Grimage shortcuts. *)

let blank = 0;; (* = Graphics.black *)

let pixmap_make (w, h) c = Umatrix.to_array_array (Umatrix.make h w c);;

let pixmap_draw pxm (x, y) =
  Pixmap.draw_pixmap pxm (Nat.to_int x) (Nat.to_int y)
;;

open Configuration_types;;
open Image;;
open Display_types;;
open Display_types_print;;
open Display_window;;

let mk_pq_alt _fake _compound full _gk = full;;

let mk_oq_alt fake _compound full = function
  | (Param_quantity, _) -> fake
  | (All_quantities, _) -> full
;;

let mk_pq_drw _fake compound full = function
  | (_, Compound_channel) -> compound
  | (_, All_channels) -> full
;;

let mk_oq_drw fake compound full = function
  | (Param_quantity, _) -> fake
  | (All_quantities, Compound_channel) -> compound
  | (All_quantities, All_channels) -> full
;;

let mk_rgb_compound cmp rgb = { rgb = rgb; compound = cmp; };;

let fk_rgb fk = Triple.make fk;;

let assert_false = (fun _ -> assert false);;
let assert_false_rgb = fk_rgb assert_false;;

let fake_rc fk = mk_rgb_compound fk (fk_rgb fk);;
let compound_rc fk cmp = mk_rgb_compound cmp (fk_rgb fk);;
let full_rc cmp rgb = mk_rgb_compound cmp rgb;;

let mk_rc_xq mk_xq fk cmp rgb gk =
  let fake = fake_rc fk in
  let compound = compound_rc fk cmp in
  let full = full_rc cmp rgb in
  mk_xq fake compound full gk
;;

let mk_rc mk_pq mk_oq = function
  | Param -> mk_rc_xq mk_pq
  | Data -> mk_rc_xq mk_oq
  | Grad -> mk_rc_xq mk_oq
;;

let mk_rc_alt fk = mk_rc mk_pq_alt mk_oq_alt fk
and mk_rc_drw fk = mk_rc mk_pq_drw mk_oq_drw fk
;;

let get_rgb rgb = Triple.map_rails rgb ((), (), ());;

let get_rgb_compound { rgb = rgb; compound = cmp; } =
  mk_rgb_compound (cmp ()) (get_rgb rgb)
;;

(* Get_pixmaps.
 For late enough evaluation of its components, it is easier to only consider
 functional rgb_compounds. If necessary, thunks are added, e.g. for pixmaps
 and lower_left_corners.  *)

let fk_get_pxm = (fun () -> [||]);;

let mk_get_pxm sz = (fun () -> pixmap_make sz blank);;
let mk_get_pxm_rgb sz = Triple.make (mk_get_pxm sz);;

let mk_get_pxms q gk sz =
  let fk = fk_get_pxm in
  let cmp = mk_get_pxm sz in
  let rgb = mk_get_pxm_rgb sz in
  mk_rc_alt  q fk cmp rgb gk
;;

let make_pixmaps q gk sz = get_rgb_compound (mk_get_pxms q gk sz);;

(* Lower_left_corners. See get_pixmaps above. *)

let fk_get_ll = (fun () -> (_0_, _0_));;

let mk_get_ll (nx, ny) m ax bx ay by =
  let two_ms = Nat.mul _2_ m in
  let sub_hor = Nat.add nx two_ms in
  let sub_ver = Nat.add ny two_ms in
  (fun () -> (natural_axpby ax sub_hor bx m, natural_axpby ay sub_ver by m))
;;

let mk_get_ll_cmp_p gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((Param_quantity, Compound_channel), _) ->
      mk_get_ll _0_ _1_ _0_ _1_
  | ((Param_quantity, All_channels), Rgb_Horizontal) ->
      mk_get_ll _3_ _3_ _0_ _1_
  | ((Param_quantity, All_channels), Rgb_Vertical) ->
      mk_get_ll _0_ _1_ _0_ _1_
  | ((All_quantities, Compound_channel), Rgb_Horizontal) ->
      mk_get_ll _1_ _3_ _0_ _1_
  | ((All_quantities, Compound_channel), Rgb_Vertical) ->
      mk_get_ll _0_ _1_ _1_ _1_
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      mk_get_ll _3_ _3_ _1_ _1_
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      mk_get_ll _1_ _3_ _0_ _1_
;;

let mk_get_ll_cmp_d gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((Param_quantity, _), _) -> assert_false
  | ((All_quantities, Compound_channel), Rgb_Horizontal) ->
      mk_get_ll _0_ _1_ _0_ _1_
  | ((All_quantities, Compound_channel), Rgb_Vertical) ->
      mk_get_ll _0_ _1_ _2_ _3_
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      mk_get_ll _3_ _3_ _2_ _3_
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      mk_get_ll _0_ _1_ _0_ _1_
;;

let mk_get_ll_cmp_g gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((Param_quantity, _), _) -> assert_false
  | ((All_quantities, Compound_channel), Rgb_Horizontal) ->
      mk_get_ll _2_ _3_ _0_ _1_
  | ((All_quantities, Compound_channel), Rgb_Vertical) ->
      mk_get_ll _0_ _1_ _0_ _1_
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      mk_get_ll _3_ _3_ _0_ _1_
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      mk_get_ll _2_ _3_ _0_ _1_
;;

let mk_get_ll_cmp = function
  | Param -> mk_get_ll_cmp_p
  | Data -> mk_get_ll_cmp_d
  | Grad -> mk_get_ll_cmp_g
;;

let mk_get_ll_rgb_p gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((_, Compound_channel), _) -> assert_false_rgb
  | ((Param_quantity, All_channels), Rgb_Horizontal) ->
      (mk_get_ll _0_ _1_ _0_ _1_,
       mk_get_ll _1_ _1_ _0_ _1_,
       mk_get_ll _2_ _1_ _0_ _1_)
  | ((Param_quantity, All_channels), Rgb_Vertical) ->
      (mk_get_ll _0_ _1_ _3_ _3_,
       mk_get_ll _0_ _1_ _2_ _3_,
       mk_get_ll _0_ _1_ _1_ _3_)
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      (mk_get_ll _0_ _1_ _1_ _1_,
       mk_get_ll _1_ _1_ _1_ _1_,
       mk_get_ll _2_ _1_ _1_ _1_)
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      (mk_get_ll _1_ _3_ _3_ _3_,
       mk_get_ll _1_ _3_ _2_ _3_,
       mk_get_ll _1_ _3_ _1_ _3_)
;;

let mk_get_ll_rgb_d gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((Param_quantity, _), _)
  | ((All_quantities, Compound_channel), _) -> assert_false_rgb
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      (mk_get_ll _0_ _1_ _2_ _3_,
       mk_get_ll _1_ _1_ _2_ _3_,
       mk_get_ll _2_ _1_ _2_ _3_)
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      (mk_get_ll _0_ _1_ _3_ _3_,
       mk_get_ll _0_ _1_ _2_ _3_,
       mk_get_ll _0_ _1_ _1_ _3_)
;;

let mk_get_ll_rgb_g gk rgb_dir sz m =
  let mk_get_ll = mk_get_ll sz m in
  match (gk, rgb_dir) with
  | ((Param_quantity, _), _)
  | ((All_quantities, Compound_channel), _) -> assert_false_rgb
  | ((All_quantities, All_channels), Rgb_Horizontal) ->
      (mk_get_ll _0_ _1_ _0_ _1_,
       mk_get_ll _1_ _1_ _0_ _1_,
       mk_get_ll _2_ _1_ _0_ _1_)
  | ((All_quantities, All_channels), Rgb_Vertical) ->
      (mk_get_ll _2_ _3_ _3_ _3_,
       mk_get_ll _2_ _3_ _2_ _3_,
       mk_get_ll _2_ _3_ _1_ _3_)
;;

let mk_get_ll_rgb = function
  | Param -> mk_get_ll_rgb_p
  | Data -> mk_get_ll_rgb_d
  | Grad -> mk_get_ll_rgb_g
;;

let mk_get_lls q gk rgb_dir sz m =
  let fk = fk_get_ll in
  let cmp = mk_get_ll_cmp q gk rgb_dir sz m in
  let rgb = mk_get_ll_rgb q gk rgb_dir sz m in
  mk_rc_drw q fk cmp rgb gk
;;

let make_lower_left_corners q gk rgb_dir sz m =
  get_rgb_compound (mk_get_lls q gk rgb_dir sz m)
;;

(* Print_pixmaps. *)

let fk_prt = (fun _ppf -> ());;

let fake_print_pixmaps = fake_rc fk_prt;;

let mk_prt q s pxm =
  (fun ppf ->
     Format.fprintf ppf "@[<hv 2>%a %s pxm =@ %a@]"
       print_quantity q
       s
       print_pixmap pxm)
;;

let mk_prt_rgb q = Triple.map2 (mk_prt q) ("red", "green", "blue");;

let make_print_pixmaps q pxms gk =
  let fk = fk_prt in
  let cmp = mk_prt q "compound" pxms.compound in
  let rgb = mk_prt_rgb q pxms.rgb in
  mk_rc_alt q fk cmp rgb gk
;;

(* Alter_pixmaps. *)

let set_pxm_pxl pxm c i j = pxm.(i).(j) <- c;;

let combine_pixmaps (pxm_r, pxm_g, pxm_b) pxm_cmp =
  for i = 0 to Array.length pxm_cmp - 1 do
    let pxm_cmp_i = pxm_cmp.(i) in
    for j = 0 to Array.length pxm_cmp_i - 1 do
      let c = pxm_r.(i).(j) + pxm_g.(i).(j) + pxm_b.(i).(j) in
      set_pxm_pxl pxm_cmp c i j;
    done;
  done;
;;

(* Function scale is the multiplication by a rational number scaling factor.
 It transforms coordinates in the computed image into coordinates in the
 displayed image. Therefore, the rational coefficient is greater than 1 when
 the displayed image is larger than the computed image, and of course it is
 lesser than 1 otherwise. *)

let set_image_pixel set_pixel =
  let { scaling_factor = factor; _; } = get_graphics_elements () in
  let scale = Prational.scale factor in
  (fun im_i im_j ->
     let imin = Nat.to_int (scale im_i)
     and imax = pred (Nat.to_int (scale (Nat.succ im_i))) in
     let jmin = Nat.to_int (scale im_j)
     and jmax = pred (Nat.to_int (scale (Nat.succ im_j))) in
     for i = imin to imax do
       for j = jmin to jmax do
         set_pixel i j;
       done;
     done;)
;;

let set_zone_pixel set_pixel stc images dom idx _uidx =
  let x =
    match images with
    | Mapping.Images ims -> ims.(Uindex.to_int idx)
    | Mapping.Constant m -> m in
  let c = stc x in
  let set_pixel = set_pixel c in
  let (im_i, im_j) = Set.element dom idx in
  set_image_pixel set_pixel im_i im_j;
;;

let set_zone pxm stc f _zn map =
  let { Mapping.domain = dom; images = images; } = f (param map) in
  let set_pixel = set_pxm_pxl pxm in
  let set_zone_pixel = set_zone_pixel set_pixel stc images dom in
  Set.iter set_zone_pixel dom;
;;

let alter_zones iter pxm stc f pim =
  let set_zone = set_zone pxm stc f in
  iter set_zone pim;
;;

let fk_alt = (fun _iter _stc _gdata _pim -> ());;

(* Tekno mk_rgb_compound does not work for polymorphic function inputs. *)
let fake_alter_pixmaps =
  {
   rgb = (fk_alt, fk_alt, fk_alt);
   compound = fk_alt;
  }
;;

let mk_alt_cmp pxms =
  (fun _iter _stc _gdata _pim -> combine_pixmaps pxms.rgb pxms.compound)
;;

let mk_alt cf sc pxm =
  (fun iter stc gdata pim ->
     let stc = stc cf in
     let f = sub_gimage sc gdata in
     alter_zones iter pxm stc f pim)
;;

let mk_alt_rgb cf sc pxms = Triple.map (mk_alt cf sc) pxms.rgb;;

let make_alter_pixmaps q cf sc pxms gk =
  let fk = fk_alt in
  let cmp = mk_alt_cmp pxms in
  let rgb = mk_alt_rgb cf sc pxms in
  mk_rc_alt q fk cmp rgb gk
;;

(* Draw_pixmaps. *)

let fk_drw = (fun () -> ());;

let fake_draw_pixmaps = fake_rc fk_drw;;

let mk_drw pxm ll = (fun () -> pixmap_draw pxm ll);;
let mk_drw_rgb = Triple.map2 mk_drw;;

let make_draw_pixmaps q pxms lls gk =
  let fk = fk_drw in
  let cmp = mk_drw pxms.compound lls.compound in
  let rgb = mk_drw_rgb pxms.rgb lls.rgb in
  mk_rc_drw q fk cmp rgb gk
;;

(* Application. *)

let apply_rgb apply rc = Triple.iter apply rc.rgb;;
let apply_rc apply rc = apply_rgb apply rc; apply rc.compound;;

let print_apply print ppf = apply_rc (fun f -> assert (print f ppf))
and draw_apply = apply_rc (fun f -> f ())
;;

let alter_apply_rgb iter stct gdatat pimt rc =
  let rgb = Triple.map (fun f -> f iter) rc.rgb in
  Triple.iter_rails3 rgb stct gdatat pimt;
;;

let alter_apply_cmp iter (stc_r, _, _) (_, gdata_g, _) (_, _, pim_b) rc =
  rc.compound iter stc_r gdata_g pim_b;
;;

let alter_apply iter stct gdatat pimt rc =
  alter_apply_rgb iter stct gdatat pimt rc;
  alter_apply_cmp iter stct gdatat pimt rc;
;;
