(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ucommand_option.mli,v 1.36 2014/01/10 15:28:04 fclement Exp $ *)

(** {3 Handle options for the command line of a program}
 
 Regular command options are the association of an option name to use on
 the command line, an action to be executed when parsing the command line,
 and a documentation to be printed for help.
 Generalized command options have the additional capability to specify
 synonyms to the option name.
 
 Command options are classified either as user, user report and debug,
 library, library report and debug, and service command options.
 This classification drives the appearance of option documentations in the
 help message: user command options come first (regular, then report and
 debug), then library command options (regular, then report and debug), and
 finally service command options.
 
 User command options are devoted to standalone programs.
 Library command options are devoted to libraries.
 
 User and library report command options are used to trigger the collection
 and printing of information during the execution of code of a program or a
 library (e.g. to monitor algorithms).
 
 User and library debug command options are used to trigger the trace of the
 execution of code of a program or a library when its behavior is not
 satisfactory.
 
 Service command options are automatically provided; they are used to trigger
 either all report or debug command options, or the printing of the help
 message that contains the documentation of all the command options.
 Option names "-report" (with aliases "--report" and "-r"), "-debug" (with
 aliases "--debug" and "-d"), and "-help" (with aliases "--help" and "-h") are
 automatically provided.
 
 The parsing of command line options may be split into phases, see the
 dedicated section.
 
 This module is implemented over module Arg of the OCaml standard library.
 Module {!Uaref} (from the OCaml user's additional library Ulib) provides a
 collection of abstract references that may be used to define the actions
 to be executed when parsing the command line.
 Module {!Udebug} (also from Ulib) provides shortcuts for the definition of
 reporting and degugging facilities, as well as error message printing. *)

(** {6 Defining command options} *)

type regular_command_option =
  {
   key : Arg.key;
   spec : Arg.spec;
   doc : Arg.doc;
  }
;;
(** The type of regular command options.
 For a given regular command option [opt], [opt.key] is the option name
 (a string starting with a dash (['-']) character), [opt.spec] is the action,
 and [opt.doc] is the documentation (a string).
 This type is equivalent to the [key * spec * doc] triples used in module
 Arg of the standard library. *)

type t;;
(** The abstract type for command options.
 The notion of command option generalizes the notion of regular command
 option by adding the possibility to specify aliases to the main option name.
 Aliases maybe of three kinds: Gnu key aliases are supposed to be long
 descriptive option names starting with two dash (['-']) characters;
 short key aliases are supposed to be short option names with a small number
 of characters (ideally one or two, plus the initial dash (['-']) character);
 other key aliases have free format.
 Gnu key and short key aliases may be computed automatically from the main
 option name (which is then supposed to be a long descriptive option name). *)

val of_regular : regular_command_option -> t;;
(** [of_regular rco] returns the command option with [rco.key] as the main
 option name, the same action and documentation, and with no key alias. *)

val with_gnu_key : Arg.key -> t -> t;;
(** [with_gnu_key k co] returns the same command option with its Gnu key set
 to [k]. *)

val with_short_key : Arg.key -> t -> t;;
(** [with_short_key k co] returns the same command option with its short key
 set to [k]. *)

val with_key_aliases : Arg.key list -> t -> t;;
(** [with_short_key kl co] returns the same command option with the elements
 of the list [kl] as key aliases. *)

val with_auto_gnu_key : t -> t;;
(** [with_auto_gnu_key co] returns the same command option with its Gnu key
 automatically set to its main key prefixed by an additional dash (['-'])
 character. *)

val with_auto_short_key : t -> t;;
(** [with_auto_short_key co] returns the same command option with its short
 key automatically set to the concatenation of initials of words of its main
 key prefixed by "-". Words in the main key are supposed to be separated by
 dash (['-']) characters. *)

(** {6 Managing command options} *)

val user_add : t -> unit;;
val library_add : t -> unit;;
val report_user_add : t -> unit;;
val report_library_add : t -> unit;;
val debug_user_add : t -> unit;;
val debug_library_add : t -> unit;;
(** Add the command option given in argument in the corresponding class of
 command options.
 For each class (user, library, user and library report, and user and library
 debug), the order of specification in the code of the program is preserved in
 the help message. *)

val user_add_all : t list -> unit;;
val library_add_all : t list -> unit;;
val report_user_add_all : t list -> unit;;
val report_library_add_all : t list -> unit;;
val debug_user_add_all : t list -> unit;;
val debug_library_add_all : t list -> unit;;
(** Similar to previous versions with no [_all] suffix, but add all the
 elements of the list of command options given in argument. *)

val clear_all : unit -> unit;;
(** Clear all already added entries of all classes of command options. *)

val user_clear : unit -> unit;;
val library_clear : unit -> unit;;
val report_user_clear : unit -> unit;;
val report_library_clear : unit -> unit;;
val debug_user_clear : unit -> unit;;
val debug_library_clear : unit -> unit;;
(** Clear already added entries in the corresponding class of command
 options. *)

val user_clear_key : Arg.key -> unit;;
val library_clear_key : Arg.key -> unit;;
val report_user_clear_key : Arg.key -> unit;;
val report_library_clear_key : Arg.key -> unit;;
val debug_user_clear_key : Arg.key -> unit;;
val debug_library_clear_key : Arg.key -> unit;;
(** Clear all already added entries in the corresponding class of command
 options which main option name matches the given [key]. *)

(** {6 Parsing command options} *)

type command_line = string array;;
(** The type of command lines. *)

val set_anon_fun : Arg.anon_fun -> unit;;
(** Set the function called on anonymous arguments.
 Default value is [ignore]. *)

val set_usage_msg : Arg.usage_msg -> unit;;
(** Set the usage message displayed by [-help] option.
 Default value is ["\nUsage: <command> [options]\n\nOptions are:"] where
 [<command>] is the command name used to invoke the program. *)

val parse : unit -> unit;;
(** [parse ()] parses the command line [Sys.argv] to treat command options.
 Anonymous function and usage message are those set by {!set_anon_fun} and
 {!set_usage_msg}.
 See module [Arg] of the standard library. *)

val parse_argv : command_line -> unit;;
(** [parse_argv argv] parses the command line [argv] to treat command options.
 See {!parse} for details. *)

(** {6 Phased parsing of command line options} *)

(** User and library command line options may be grouped into phases;
 then, command line options may be parsed successively phase by phase.
 This way, some mandatory options may be processed first to be able to further
 process the remaining options.
 For instance, this may be used to obtain an early load of the configuration of
 an iterative algorithm when proceeding to a warm restart.
 
 To add {e phased options} to phase named ["foobar"], use either {!phase_add},
 or {!phase_add_all}, with [Phase "foobar"] as first argument.
 Of course, phased options of a library should be clearly documented to allow
 for a proper phased parsing by the user;
 in particular, the name of the phases should be clearly mentioned.
 
 To process all phased options of phase ["foobar"] in the command line
 arguments [Sys.argv] (resp. [argv]), use function {!parse_phase} (resp.
 {!parse_phase_argv}), with [Phase "foobar"] as first argument.
 The remaining command line arguments in [Sys.argv] (resp. [argv]) may be
 further parsed either again with {!parse_phase} (resp. {!parse_phase_argv})
 to process options of another phase, or with {!parse} (resp. {!parse_argv})
 to process the whole remainder.
 
 Note that the process of anonymous arguments cannot be phased, thus anonymous
 arguments can only be parsed with {!parse} and {!parse_argv} functions. *)

type phase_name = Phase of string;;

val phase_add : phase_name -> Arg.key -> unit;;
(** Add the command option associated to the key argument to the specified
 phase.

 Raise [Failure] if the key is not recorded in any of the command option
 classes. *)

val phase_add_all : phase_name -> Arg.key list -> unit;;
(** Add the command options associated to the keys argument to the specified
 phase.

 Raise [Failure] if any of the keys is not recorded in any of the command
 option classes. *)

val parse_phase : phase_name -> unit;;
(** [parse_phase phase_name] parses the command line [Sys.argv] to treat
 phased options from phase [phase_name]. *)

val parse_phase_argv : phase_name -> command_line -> command_line;;
(** [parse_phase_argv phase_name argv] parses the command line arguments
 [argv] to treat phased options from phase [phase_name], and returns the
 remaining command line arguments. *)
