(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: udebug.mli,v 1.24 2014/01/10 15:28:04 fclement Exp $ *)

(** {3 Report/debug/error option flags and printing facilities}
 
 This module provides functionalities related to the set up of specific
 reporting and debugging option flags (either marked for [User] of [Library]
 purpose) to be used on the command line, and specific printing facilities for
 reporting/monitoring, debugging, and printing error messages.
 
 The messages are printed on stderr.
 The form of the messages is as follows where tags may be static or dynamic:
   - reporting: "<tag> <message>"
   - debugging: "<tag> <function name>: <message>"
   - error:     "<tag> Fatal error in <function name>: <message>"
 
 This module is based on module {!Ucommand_option} (from the OCaml user's
 additional library Ulib). *)

(** {6 Messages with no tags} *)

type purpose = | User | Library
and software_name = string
and prefix = string
;;
(** Basic types for [Udebug]. *)

type short_key_prefix =
  | Automatic
  | Short of prefix

and key_prefixes =
  {
   long : prefix;
   short : short_key_prefix;
  }
;;
(** The type of key prefixes.
   Long and short prefixes to build dedicated reporting and debugging option
   flags to be used on the command line.
   When the [short] field equals [Automatic], the short key prefix is
   eventually computed from the initials of the words of the [long] field
   (where the separator is the dash character ['-']). *)

type 'a printing_facilities =
  {
   report : Upervasives.message -> bool;
   report_action : (unit -> Upervasives.message) -> bool;
   debug : Upervasives.function_name -> Upervasives.message -> bool;
   debug_action :
     Upervasives.function_name -> (unit -> Upervasives.message) -> bool;
   fatal_error :
     Upervasives.function_name -> Upervasives.message -> int -> 'a;
  }
;;
(** The type of printing facilities.
   Fields [report] and [debug] are printing functions that return the value
   [true].
   Fields [report_action] and [debug_action] are similar to [report] and
   [debug] but their last argument is an action that generates the message to
   be displayed.
   Field [fatal_error] is a printing function that terminates with its third
   argument as status code.
   
   Note that the boolean result of reporting and debugging functions make them
   suitable to be used in [assert] expressions;
   that way, all the reporting/debugging machinery of a program can be
   bypassed with a compilation with the {e -noassert} option. *)

val make_printing_facilities :
  purpose -> software_name -> key_prefixes -> 'a printing_facilities
;;
(** Printing facilities construction function with no tag prefix.
   Reporting and debugging option flags to be used on the command line are
   automatically computed from the key prefixes. *)

val make_key_prefixes : software_name -> key_prefixes;;
(** Key prefixes construction function.
   The long key prefix is a stripped version of the software name (lowercase,
   all consecutive non alphanumeric characters are replaced by a single dash
   (['-']) character, and characters from the ISO Latin-1 (8859-1) character
   set are mapped to the US-ASCII character set).
   The short key prefix option is [Automatic] (see type {!key_prefixes}). *)

(** {6 Messages with static tags} *)

val make_printing_facilities_with_static_tag :
  purpose -> software_name -> key_prefixes -> 'a printing_facilities
;;
(** Printing facilities construction function with static tag prefix.
   It is similar to the printing facilities construction function
   {!make_printing_facilities} except that a static tag is prefixed to each
   printed message. This tag is the software name between square brackets
   characters (['['], [']']). *)

(** {6 Messages with dynamic tags} *)

type tag = string
and 'a dynamic_tag = 'a -> tag
;;
(** Dynamic tags are functions returning a string. *)

type ('a, 'b) dynamic_printing_facilities =
  {
   dyn_report : 'a -> Upervasives.message -> bool;
   dyn_report_action : 'a -> (unit -> Upervasives.message) -> bool;
   dyn_debug :
     'a -> Upervasives.function_name -> Upervasives.message -> bool;
   dyn_debug_action :
     'a -> Upervasives.function_name -> (unit -> Upervasives.message) -> bool;
   dyn_fatal_error :
     'a -> Upervasives.function_name -> Upervasives.message -> int -> 'b;
  }
;;
(** The type of dynamic printing facilities.
   The fields are the similar to those of the type {!printing_facilities}
   except that they have an additional first argument to compute the dynamic
   tag. *)

val make_printing_facilities_with_dynamic_tags :
  purpose -> software_name -> key_prefixes -> 'a dynamic_tag ->
  ('a, 'b) dynamic_printing_facilities
;;
(** Printing facilities construction function with dynamic tag prefixes.
   It is similar to the (static) printing facilities construction function
   {!make_printing_facilities_with_static_tag} except that the tag generator
   is given as an additional third argument. *)
