(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: upervasives.ml,v 1.32 2014/01/10 15:28:04 fclement Exp $ *)

(** Extensions to the initially opened module. *)

(* Error messages. *)
type module_name = string
and function_name = string
and message = string
;;

let error_message mod_name fun_name msg =
  let fun_name =
    if String.length mod_name = 0 then fun_name else
    if String.length fun_name = 0 then mod_name else
    Printf.sprintf "%s.%s" mod_name fun_name in
  if String.length msg = 0 then Printf.sprintf "%s" fun_name else
  if String.length fun_name = 0 then Printf.sprintf "%s" msg else
  Printf.sprintf "%s: %s" fun_name msg
;;

(* Exceptions. *)
let invalid_arg mod_name fun_name msg =
  Pervasives.invalid_arg (error_message mod_name fun_name msg)
;;

let failwith mod_name fun_name msg =
  Pervasives.failwith (error_message mod_name fun_name msg)
;;

(* Functional helpers. *)

external id : 'a -> 'a = "%identity";;

let compose g f x = g (f x);;

let proj2_1 x _ = x;;
let proj2_2 _ y = y;;

let option f d = function
  | Some x -> f x
  | None -> d
;;

let lifto f = option (fun x -> Some (f x)) None;;

(* Integer arithmetic. *)

type index = Unatural.t;;

let index_of_int = Unatural.of_int;;
let index_to_int = Unatural.to_int;;

type length = Unatural.t;;

let length_of_int = Unatural.of_int;;
let length_to_int = Unatural.to_int;;

(* Floating-point arithmetic. *)

let round_pos x = int_of_float (x +. 0.5);;
let round_neg x = int_of_float (x -. 0.5);;
let round x = if x >= 0. then round_pos x else round_neg x;;

(* Conversion. *)

let char_of_int64 i = char_of_int (Int64.to_int (Int64.logand i 0xffL));;

let string_init n f =
  if n = 0 then "" else
  let res =
    try String.make n (f 0) with
    | Invalid_argument _ -> Pervasives.invalid_arg "string_init" in
  for i = 1 to pred n do
    res.[i] <- f i;
  done;
  res
;;

let bit_string_of_float f =
  let i = Int64.bits_of_float f in
  string_init 8 (fun j -> char_of_int64 (Int64.shift_right i (8*j)))
;;

let int64_of_char c = Int64.of_int (int_of_char c);;

let float_of_bit_string s =
  assert (String.length s = 8);
  let rec loop accu j =
    if j < 0 then accu else
    loop (Int64.add (Int64.shift_left accu 8) (int64_of_char s.[j])) (j - 1) in
  Int64.float_of_bits (loop (int64_of_char s.[7]) 6)
;;

(* Input/output. *)

let pr_unit pr_string () = pr_string "()";;

let print_unit = pr_unit print_string;;

let prerr_unit = pr_unit prerr_string;;

let pr_bool pr_string b =
  if b then pr_string "true" else pr_string "false"
;;

let print_bool = pr_bool print_string;;

let prerr_bool = pr_bool prerr_string;;
  

let open_append fname =
  let mode = [Open_wronly; Open_append; Open_creat; Open_text;] in
  let perm = 0o644 in
  open_out_gen mode perm fname
;;

let open_append_bin fname =
  let mode = [Open_wronly; Open_append; Open_creat; Open_binary;] in
  let perm = 0o644 in
  open_out_gen mode perm fname
;;
